#include "../../src/ErrorSimulator.hpp"
#include "../Tests.hpp"
#include "core/data/ProbingHashMap.hpp"
#include "core/utils/Error.hpp"

template struct Core::ProbingHashMap<int, int>;
using IntMap = Core::ProbingHashMap<int, int>;

static void testAdd() {
    IntMap map;
    map.add(5, 4);
    int* value = map.search(5);
    CORE_TEST_NOT_NULL(value);
    if(value != nullptr) {
        CORE_TEST_EQUAL(4, *value);
    }
}

static void testMultipleAdd() {
    IntMap map;
    map.add(5, 4).add(10, 3).add(15, 2);
    CORE_TEST_TRUE(map.contains(5));
    CORE_TEST_TRUE(map.contains(10));
    CORE_TEST_TRUE(map.contains(15));
    int* a = map.search(5);
    int* b = map.search(10);
    int* c = map.search(15);
    CORE_TEST_NOT_NULL(a);
    CORE_TEST_NOT_NULL(b);
    CORE_TEST_NOT_NULL(c);
    if(a != nullptr && b != nullptr && c != nullptr) {
        CORE_TEST_EQUAL(4, *a);
        CORE_TEST_EQUAL(3, *b);
        CORE_TEST_EQUAL(2, *c);
    }
}

static void testSearch() {
    IntMap map;
    CORE_TEST_NULL(map.search(6));
    map.add(5, 4).add(10, 3).add(15, 2);
    CORE_TEST_NULL(map.search(6));
}

static void testAddReplace() {
    IntMap map;
    map.add(5, 4).add(5, 10);
    CORE_TEST_TRUE(map.contains(5));
    int* a = map.search(5);
    CORE_TEST_NOT_NULL(a);
    if(a != nullptr) {
        CORE_TEST_EQUAL(10, *a);
    }
}

static void testClear() {
    IntMap map;
    map.add(5, 4).add(4, 10);
    map.clear();
    CORE_TEST_FALSE(map.contains(5));
    CORE_TEST_FALSE(map.contains(4));
}

static void testOverflow(bool light) {
    int limit = light ? 10000 : 100000;
    IntMap map;
    for(int i = 0; i < limit; i++) {
        map.add(i, i);
    }
    for(int i = 0; i < limit; i++) {
        CORE_TEST_TRUE(map.contains(i));
    }
}

static int aInstances = 0;

struct ProbingHashMapTestStruct final {
    int a;
    int b;

    ProbingHashMapTestStruct(int a_, int b_) : a(a_), b(b_) {
        aInstances++;
    }

    ProbingHashMapTestStruct(const ProbingHashMapTestStruct& o)
        : a(o.a), b(o.b) {
        aInstances++;
    }

    ProbingHashMapTestStruct(ProbingHashMapTestStruct&& o) : a(o.a), b(o.b) {
        aInstances++;
    }

    ~ProbingHashMapTestStruct() {
        aInstances--;
    }

    ProbingHashMapTestStruct&
    operator=(const ProbingHashMapTestStruct& o) = default;
    ProbingHashMapTestStruct& operator=(ProbingHashMapTestStruct&& o) = default;

    bool operator==(const ProbingHashMapTestStruct& other) const {
        return a == other.a && b == other.b;
    }

    template<typename String>
    check_return Core::Error toString(String& s) const {
        CORE_RETURN_ERROR(s.append("A("));
        CORE_RETURN_ERROR(s.append(a));
        CORE_RETURN_ERROR(s.append(", "));
        CORE_RETURN_ERROR(s.append(b));
        CORE_RETURN_ERROR(s.append(")"));
        return Core::ErrorCode::NONE;
    }
};

static void testEmplace() {
    {
        Core::ProbingHashMap<int, ProbingHashMapTestStruct> map;

        ProbingHashMapTestStruct* ar = nullptr;
        CORE_TEST_TRUE(map.tryEmplace(ar, 0, 3, 4));
        CORE_TEST_TRUE(map.tryEmplace(ar, 3, 4, 5));
        CORE_TEST_TRUE(map.tryEmplace(ar, 20, 5, 6));
        CORE_TEST_FALSE(map.tryEmplace(ar, 3, 6, 7));
        CORE_TEST_FALSE(map.tryEmplace(ar, 20, 7, 8));

        ProbingHashMapTestStruct* a = map.search(0);
        ProbingHashMapTestStruct* b = map.search(3);
        ProbingHashMapTestStruct* c = map.search(20);

        CORE_TEST_NOT_NULL(a);
        CORE_TEST_NOT_NULL(b);
        CORE_TEST_NOT_NULL(c);

        if(a != nullptr && b != nullptr && c != nullptr) {
            CORE_TEST_EQUAL(ProbingHashMapTestStruct(3, 4), *a);
            CORE_TEST_EQUAL(ProbingHashMapTestStruct(4, 5), *b);
            CORE_TEST_EQUAL(ProbingHashMapTestStruct(5, 6), *c);
        }
    }
    CORE_TEST_EQUAL(0, aInstances);
}

static void testToString1() {
    IntMap map;
    map.add(1, 3).add(2, 4).add(3, 5);
    CORE_TEST_STRING("[2 = 4, 1 = 3, 3 = 5]", map);
}

static void testToString2() {
    IntMap map;
    map.add(1, 3);
    CORE_TEST_STRING("[1 = 3]", map);
}

static void testToString3() {
    IntMap map;
    CORE_TEST_STRING("[]", map);
}

static void testCopy() {
    IntMap map;
    map.add(1, 3).add(2, 4).add(3, 5);
    IntMap copy = map;

    int* a[6] = {map.search(1),  map.search(2),  map.search(3),
                 copy.search(1), copy.search(2), copy.search(3)};
    for(int i = 0; i < 3; i++) {
        CORE_TEST_NOT_NULL(a[i]);
        CORE_TEST_NOT_NULL(a[i + 3]);
        if(a[i] != nullptr && a[i + 3] != nullptr) {
            CORE_TEST_EQUAL(*(a[i]), *(a[i + 3]));
        }
    }
}

static void testMove() {
    IntMap map;
    map.add(1, 3).add(2, 4).add(3, 5);
    IntMap move(Core::move(map));

    int* a = move.search(1);
    int* b = move.search(2);
    int* c = move.search(3);

    CORE_TEST_NOT_NULL(a);
    CORE_TEST_NOT_NULL(b);
    CORE_TEST_NOT_NULL(c);

    if(a != nullptr && b != nullptr && c != nullptr) {
        CORE_TEST_EQUAL(3, *a);
        CORE_TEST_EQUAL(4, *b);
        CORE_TEST_EQUAL(5, *c);
    }
}

static void testMoveAssignment() {
    IntMap map;
    map.add(1, 3).add(2, 4).add(3, 5);

    IntMap move;
    move = Core::move(map);

    int* a = move.search(1);
    int* b = move.search(2);
    int* c = move.search(3);

    CORE_TEST_NOT_NULL(a);
    CORE_TEST_NOT_NULL(b);
    CORE_TEST_NOT_NULL(c);

    if(a != nullptr && b != nullptr && c != nullptr) {
        CORE_TEST_EQUAL(3, *a);
        CORE_TEST_EQUAL(4, *b);
        CORE_TEST_EQUAL(5, *c);
    }
}

static void testEntryForEach() {
    IntMap map;
    map.add(5, 4).add(10, 3).add(15, 2);

    int counter = 0;
    for(auto entry : map) {
        counter += entry.getKey() + entry.value;
    }
    CORE_TEST_EQUAL(39, counter);

    const IntMap& cmap = map;
    counter = 0;
    for(auto entry : cmap) {
        counter += entry.getKey() + entry.value;
    }
    CORE_TEST_EQUAL(39, counter);
}

static void testKeyForEach() {
    IntMap map;
    map.add(5, 4).add(10, 3).add(15, 2);

    int counter = 0;
    for(const int& key : map.getKeys()) {
        counter += key;
    }
    CORE_TEST_EQUAL(30, counter);

    const IntMap& cmap = map;
    counter = 0;
    for(const int& key : cmap.getKeys()) {
        counter += key;
    }
    CORE_TEST_EQUAL(30, counter);
}

static void testValueForEach() {
    IntMap map;
    map.add(5, 4).add(10, 3).add(15, 2);

    int counter = 0;
    for(int& value : map.getValues()) {
        counter += value;
    }
    CORE_TEST_EQUAL(9, counter);

    const IntMap& cmap = map;
    counter = 0;
    for(const int& value : cmap.getValues()) {
        counter += value;
    }
    CORE_TEST_EQUAL(9, counter);
}

template<typename T>
static void testType() {
    Core::ProbingHashMap<T, int> m;
    m.add(T(), 3);
}

static void testTypes() {
    testType<char>();
    testType<signed char>();
    testType<signed short>();
    testType<signed int>();
    testType<signed long>();
    testType<signed long long>();
    testType<unsigned char>();
    testType<unsigned short>();
    testType<unsigned int>();
    testType<unsigned long>();
    testType<unsigned long long>();
}

static void testInsertInvalid() {
    IntMap map;
    int* v;
    CORE_TEST_TRUE(map.tryEmplace(v, Core::emptyValue<int>(), 2));
    CORE_TEST_EQUAL(3, map.put(Core::emptyValue<int>(), 3));
}

static void testAddCollisions() {
    IntMap map;
    for(int i = 0; i < 8; i++) {
        map.add(i * 16, i);
    }
}

void Core::testProbingHashMap(bool light) {
    testAdd();
    testMultipleAdd();
    testSearch();
    testAddReplace();
    testClear();
    testOverflow(light);
    testEmplace();
    testToString1();
    testToString2();
    testToString3();
    testCopy();
    testMove();
    testMoveAssignment();
    testEntryForEach();
    testKeyForEach();
    testValueForEach();
    testTypes();
    testInsertInvalid();
    testAddCollisions();
}
